import os
from collections import deque

from macholib.dyld import framework_info
from macholib.MachOGraph import MachOGraph, MissingMachO
from macholib.util import (
    flipwritable,
    has_filename_filter,
    in_system_path,
    iter_platform_files,
    mergecopy,
    mergetree,
)


class ExcludedMachO(MissingMachO):
    pass


class FilteredMachOGraph(MachOGraph):
    def __init__(self, delegate, *args, **kwargs):
        super(FilteredMachOGraph, self).__init__(*args, **kwargs)
        self.delegate = delegate

    def createNode(self, cls, name):
        cls = self.delegate.getClass(name, cls)
        res = super(FilteredMachOGraph, self).createNode(cls, name)
        return self.delegate.update_node(res)

    def locate(self, filename, loader=None):
        newname = super(FilteredMachOGraph, self).locate(filename, loader)
        if newname is None:
            return None
        return self.delegate.locate(newname, loader=loader)


class MachOStandalone(object):
    def __init__(self, base, dest=None, graph=None, env=None, executable_path=None):
        self.base = os.path.join(os.path.abspath(base), "")
        if dest is None:
            dest = os.path.join(self.base, "Contents", "Frameworks")
        self.dest = dest
        self.mm = FilteredMachOGraph(
            self, graph=graph, env=env, executable_path=executable_path
        )
        self.changemap = {}
        self.excludes = []
        self.pending = deque()

    def update_node(self, m):
        return m

    def getClass(self, name, cls):
        if in_system_path(name):
            return ExcludedMachO
        for base in self.excludes:
            if name.startswith(base):
                return ExcludedMachO
        return cls

    def locate(self, filename, loader=None):
        if in_system_path(filename):
            return filename
        if filename.startswith(self.base):
            return filename
        for base in self.excludes:
            if filename.startswith(base):
                return filename
        if filename in self.changemap:
            return self.changemap[filename]
        info = framework_info(filename)
        if info is None:
            res = self.copy_dylib(filename)
            self.changemap[filename] = res
            return res
        else:
            res = self.copy_framework(info)
            self.changemap[filename] = res
            return res

    def copy_dylib(self, filename):
        # When the filename is a symlink use the basename of the target of
        # the link as the name in standalone bundle. This avoids problems
        # when two libraries link to the same dylib but using different
        # symlinks.
        if os.path.islink(filename):
            dest = os.path.join(self.dest, os.path.basename(os.path.realpath(filename)))
        else:
            dest = os.path.join(self.dest, os.path.basename(filename))

        if not os.path.exists(dest):
            self.mergecopy(filename, dest)
        return dest

    def mergecopy(self, src, dest):
        return mergecopy(src, dest)

    def mergetree(self, src, dest):
        return mergetree(src, dest)

    def copy_framework(self, info):
        dest = os.path.join(self.dest, info["shortname"] + ".framework")
        destfn = os.path.join(self.dest, info["name"])
        src = os.path.join(info["location"], info["shortname"] + ".framework")
        if not os.path.exists(dest):
            self.mergetree(src, dest)
            self.pending.append((destfn, iter_platform_files(dest)))
        return destfn

    def run(self, platfiles=None, contents=None):
        mm = self.mm
        if contents is None:
            contents = "@executable_path/.."
        if platfiles is None:
            platfiles = iter_platform_files(self.base)

        for fn in platfiles:
            mm.run_file(fn)

        while self.pending:
            fmwk, files = self.pending.popleft()
            ref = mm.findNode(fmwk)
            for fn in files:
                mm.run_file(fn, caller=ref)

        changemap = {}
        skipcontents = os.path.join(os.path.dirname(self.dest), "")
        machfiles = []

        for node in mm.flatten(has_filename_filter):
            machfiles.append(node)
            dest = os.path.join(
                contents,
                os.path.normpath(node.filename[len(skipcontents) :]),  # noqa: E203
            )
            changemap[node.filename] = dest

        def changefunc(path):
            print("changefunc: ", path)
            if path.startswith("@loader_path/"):
                # This is a quick hack for py2app: In that
                # usecase paths like this are found in the load
                # commands of relocatable wheels. Those don't
                # need rewriting.
                return path

            elif path.startswith("@rpath/"):
                # Another hack for py2app: In most cases an
                # @rpath path doesn't require updates.
                return path

            res = mm.locate(path)
            rv = changemap.get(res)
            if rv is None and path.startswith("@loader_path/"):
                rv = changemap.get(mm.locate(mm.trans_table.get((node.filename, path))))
            if rv is None and path.startswith("@rpath/"):
                rv = changemap.get(mm.locate(mm.trans_table.get((node.filename, path))))
            return rv

        for node in machfiles:
            fn = mm.locate(node.filename)
            if fn is None:
                continue
            rewroteAny = False
            for _header in node.headers:
                if node.rewriteLoadCommands(changefunc):
                    rewroteAny = True
            if rewroteAny:
                old_mode = flipwritable(fn)
                try:
                    with open(fn, "rb+") as f:
                        for _header in node.headers:
                            f.seek(0)
                            node.write(f)
                        f.seek(0, 2)
                        f.flush()
                finally:
                    flipwritable(fn, old_mode)

        allfiles = [mm.locate(node.filename) for node in machfiles]
        return set(filter(None, allfiles))



import os
import sys

import macholib

if sys.version_info[:2] <= (2, 6):
    import unittest2 as unittest
else:
    import unittest


class TestVersion(unittest.TestCase):
    def test_version(self):
        fn = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "setup.cfg"
        )
        with open(fn) as fp:
            for ln in fp:
                if ln.startswith("version"):
                    setup_version = ln.split("=")[-1].strip()
                    break
            else:
                self.fail("Cannot determine wheel version")
        self.assertEqual(setup_version, macholib.__version__)


if __name__ == "__main__":
    unittest.main()



import mmap
import struct
import sys
import unittest

from macholib import ptypes

if sys.version_info[:2] <= (2, 6):
    import unittest2 as unittest
else:
    import unittest

try:
    from io import BytesIO
except ImportError:
    from cStringIO import StringIO as BytesIO

try:
    long
except NameError:
    long = int


class TestPTypes(unittest.TestCase):
    if not hasattr(unittest.TestCase, "assertIsSubclass"):

        def assertIsSubclass(self, class1, class2, message=None):
            self.assertTrue(
                issubclass(class1, class2),
                message or "%r is not a subclass of %r" % (class1, class2),
            )

    if not hasattr(unittest.TestCase, "assertIsInstance"):

        def assertIsInstance(self, value, types, message=None):
            self.assertTrue(
                isinstance(value, types),
                message or "%r is not an instance of %r" % (value, types),
            )

    def test_sizeof(self):
        self.assertEqual(ptypes.sizeof(b"foobar"), 6)

        self.assertRaises(ValueError, ptypes.sizeof, [])
        self.assertRaises(ValueError, ptypes.sizeof, {})
        self.assertRaises(ValueError, ptypes.sizeof, b"foo".decode("ascii"))

        class M(object):
            pass

        m = M()
        m._size_ = 42
        self.assertEqual(ptypes.sizeof(m), 42)

    def verifyType(self, ptype, size, pytype, values):
        self.assertEqual(ptypes.sizeof(ptype), size)
        self.assertIsSubclass(ptype, pytype)

        for v in values:
            pv = ptype(v)
            packed = pv.to_str()
            self.assertIsInstance(packed, bytes)
            self.assertEqual(len(packed), size)

            unp = ptype.from_str(packed)
            self.assertIsInstance(unp, ptype)
            self.assertEqual(unp, pv)

            fp = BytesIO(packed)
            unp = ptype.from_fileobj(fp)
            fp.close()
            self.assertIsInstance(unp, ptype)
            self.assertEqual(unp, pv)

            fp = BytesIO()
            pv.to_fileobj(fp)
            data = fp.getvalue()
            fp.close()
            self.assertEqual(data, packed)

            mm = mmap.mmap(-1, size + 20)
            mm[:] = b"\x00" * (size + 20)
            pv.to_mmap(mm, 10)

            self.assertEqual(ptype.from_mmap(mm, 10), pv)
            self.assertEqual(mm[:], (b"\x00" * 10) + packed + (b"\x00" * 10))

            self.assertEqual(ptype.from_tuple((v,)), pv)

    def test_basic_types(self):
        self.verifyType(ptypes.p_char, 1, bytes, [b"a", b"b"])
        self.verifyType(ptypes.p_int8, 1, int, [1, 42, -4])
        self.verifyType(ptypes.p_uint8, 1, int, [1, 42, 253])

        self.verifyType(ptypes.p_int16, 2, int, [1, 400, -10, -5000])
        self.verifyType(ptypes.p_uint16, 2, int, [1, 400, 65000])

        self.verifyType(ptypes.p_int32, 4, int, [1, 400, 2**24, -10, -5000, -(2**24)])
        self.verifyType(ptypes.p_uint32, 4, long, [1, 400, 2 * 31 + 5, 65000])

        self.verifyType(ptypes.p_int64, 8, long, [1, 400, 2**43, -10, -5000, -(2**43)])
        self.verifyType(ptypes.p_uint64, 8, long, [1, 400, 2 * 63 + 5, 65000])

        self.verifyType(ptypes.p_float, 4, float, [1.0, 42.5])
        self.verifyType(ptypes.p_double, 8, float, [1.0, 42.5])

    def test_basic_types_deprecated(self):
        self.verifyType(ptypes.p_byte, 1, int, [1, 42, -4])
        self.verifyType(ptypes.p_ubyte, 1, int, [1, 42, 253])

        self.verifyType(ptypes.p_short, 2, int, [1, 400, -10, -5000])
        self.verifyType(ptypes.p_ushort, 2, int, [1, 400, 65000])

        self.verifyType(ptypes.p_int, 4, int, [1, 400, 2**24, -10, -5000, -(2**24)])
        self.verifyType(ptypes.p_uint, 4, long, [1, 400, 2 * 31 + 5, 65000])

        self.verifyType(ptypes.p_long, 4, int, [1, 400, 2**24, -10, -5000, -(2**24)])
        self.verifyType(ptypes.p_ulong, 4, long, [1, 400, 2 * 31 + 5, 65000])

        self.verifyType(
            ptypes.p_longlong, 8, long, [1, 400, 2**43, -10, -5000, -(2**43)]
        )
        self.verifyType(ptypes.p_ulonglong, 8, long, [1, 400, 2 * 63 + 5, 65000])


class TestPTypesPrivate(unittest.TestCase):
    # These are tests for functions that aren't part of the public
    # API.

    def test_formatinfo(self):
        self.assertEqual(ptypes._formatinfo(">b"), (1, 1))
        self.assertEqual(ptypes._formatinfo(">h"), (2, 1))
        self.assertEqual(ptypes._formatinfo(">HhL"), (8, 3))
        self.assertEqual(ptypes._formatinfo("<b"), (1, 1))
        self.assertEqual(ptypes._formatinfo("<h"), (2, 1))
        self.assertEqual(ptypes._formatinfo("<HhL"), (8, 3))


class MyStructure(ptypes.Structure):
    _fields_ = (("foo", ptypes.p_int32), ("bar", ptypes.p_uint8))


class MyFunStructure(ptypes.Structure):
    _fields_ = (("fun", ptypes.p_char), ("mystruct", MyStructure))


class TestPTypesSimple(unittest.TestCase):
    # Quick port of tests that used to be part of
    # the macholib.ptypes source code
    #
    # Moving these in a structured manner to TestPTypes
    # would be nice, but is not extremely important.

    if not hasattr(unittest.TestCase, "subTest"):

        def subTest(self, msg, **kw):
            class STHelper(object):
                def __enter__(self):
                    return self

                def __exit__(self, type, value, traceback):
                    return False

            return STHelper()

    def testBasic(self):
        for endian in "><":
            kw = dict(_endian_=endian)
            MYSTRUCTURE = b"\x00\x01\x02\x03\xff"
            for fn, args in [
                ("from_str", (MYSTRUCTURE,)),
                ("from_mmap", (MYSTRUCTURE, 0)),
                ("from_fileobj", (BytesIO(MYSTRUCTURE),)),
            ]:
                with self.subTest("MYSTRUCTURE", endian=endian, fn=fn):
                    myStructure = getattr(MyStructure, fn)(*args, **kw)
                    if endian == ">":
                        self.assertEqual(myStructure.foo, 0x00010203)
                    else:
                        self.assertEqual(myStructure.foo, 0x03020100)
                    self.assertEqual(myStructure.bar, 0xFF)
                    self.assertEqual(myStructure._endian_, endian)
                    self.assertEqual(myStructure.to_str(), MYSTRUCTURE)

            MYFUNSTRUCTURE = b"!" + MYSTRUCTURE
            for fn, args in [
                ("from_str", (MYFUNSTRUCTURE,)),
                ("from_mmap", (MYFUNSTRUCTURE, 0)),
                ("from_fileobj", (BytesIO(MYFUNSTRUCTURE),)),
            ]:
                with self.subTest("MYFUNSTRUCTURE", fn=fn):
                    myFunStructure = getattr(MyFunStructure, fn)(*args, **kw)
                    self.assertEqual(myFunStructure.mystruct, myStructure)
                    self.assertEqual(
                        myFunStructure.fun, b"!", (myFunStructure.fun, b"!")
                    )
                    self.assertEqual(myFunStructure.to_str(), MYFUNSTRUCTURE)

            sio = BytesIO()
            myFunStructure.to_fileobj(sio)
            self.assertEqual(sio.getvalue(), MYFUNSTRUCTURE)

            mm = mmap.mmap(-1, ptypes.sizeof(MyFunStructure) * 2)
            mm[:] = b"\x00" * (ptypes.sizeof(MyFunStructure) * 2)
            myFunStructure.to_mmap(mm, 0)
            self.assertEqual(MyFunStructure.from_mmap(mm, 0, **kw), myFunStructure)
            self.assertEqual(mm[: ptypes.sizeof(MyFunStructure)], MYFUNSTRUCTURE)
            self.assertEqual(
                mm[ptypes.sizeof(MyFunStructure) :],
                b"\x00" * ptypes.sizeof(MyFunStructure),
            )
            myFunStructure.to_mmap(mm, ptypes.sizeof(MyFunStructure))
            self.assertEqual(mm[:], MYFUNSTRUCTURE + MYFUNSTRUCTURE)
            self.assertEqual(
                MyFunStructure.from_mmap(mm, ptypes.sizeof(MyFunStructure), **kw),
                myFunStructure,
            )

    def test_issue14(self):
        raw = b"\x01\x00\x00\x00\x00\x00\x00\x00"
        v = ptypes.p_uint64.from_str(raw, _endian_="<")
        self.assertEqual(v, struct.unpack("<q", raw)[0])

        v = ptypes.p_uint64.from_str(raw, _endian_=">")
        self.assertEqual(v, struct.unpack(">q", raw)[0])


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import sys

from macholib import _cmdline, macho_dump, macho_find, util

if sys.version_info[:2] <= (2, 6):
    import unittest2 as unittest
else:
    import unittest

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class TestCmdLine(unittest.TestCase):
    # This test is no longer valid:
    def no_test_main_is_shared(self):
        self.assertTrue(macho_dump.main is _cmdline.main)
        self.assertTrue(macho_find.main is _cmdline.main)

    def test_check_file(self):
        record = []

        def record_cb(fp, path):
            record.append((fp, path))

        self.assertEqual(_cmdline.check_file(sys.stdout, "/bin/sh", record_cb), 0)
        self.assertEqual(record, [(sys.stdout, "/bin/sh")])

        saved_stderr = sys.stderr
        saved_argv = sys.argv
        try:
            sys.stderr = StringIO()
            sys.argv = ["macho_test"]

            record[:] = []
            self.assertEqual(
                _cmdline.check_file(sys.stdout, "/bin/no-shell", record_cb), 1
            )
            self.assertEqual(record, [])
            self.assertEqual(
                sys.stderr.getvalue(),
                "macho_test: /bin/no-shell: No such file or directory\n",
            )
            self.assertEqual(record, [])

            shutil.copy("/bin/sh", "test.exec")
            os.chmod("test.exec", 0)

            sys.stderr = StringIO()
            self.assertEqual(_cmdline.check_file(sys.stdout, "test.exec", record_cb), 1)
            self.assertEqual(record, [])
            self.assertEqual(
                sys.stderr.getvalue(),
                "macho_test: test.exec: [Errno 13] Permission denied: 'test.exec'\n",
            )
            self.assertEqual(record, [])

        finally:
            sys.stderr = sys.stderr
            sys.argv = saved_argv
            if os.path.exists("test.exec"):
                os.unlink("test.exec")

    def test_shared_main(self):
        saved_stderr = sys.stderr
        saved_argv = sys.argv
        try:
            sys.stderr = StringIO()

            sys.argv = ["macho_tool"]

            self.assertEqual(_cmdline.main(lambda *args: None), 1)
            self.assertEqual(sys.stderr.getvalue(), "Usage: macho_tool filename...\n")

            names = []

            def record_names(fp, name):
                self.assertEqual(fp, sys.stdout)
                names.append(name)

            sys.stderr = StringIO()
            sys.argv = ["macho_tool", "/bin/sh"]
            self.assertEqual(_cmdline.main(record_names), 0)
            self.assertEqual(sys.stderr.getvalue(), "")
            self.assertEqual(names, ["/bin/sh"])

            names = []
            sys.stderr = StringIO()
            sys.argv = ["macho_tool", "/bin/sh", "/bin/ls"]
            self.assertEqual(_cmdline.main(record_names), 0)
            self.assertEqual(sys.stderr.getvalue(), "")
            self.assertEqual(names, ["/bin/sh", "/bin/ls"])

            names = []
            sys.stderr = StringIO()
            sys.argv = ["macho_tool", "/bin"]
            self.assertEqual(_cmdline.main(record_names), 0)
            self.assertEqual(sys.stderr.getvalue(), "")
            names.sort()
            dn = "/bin"
            real_names = [
                os.path.join(dn, fn)
                for fn in os.listdir(dn)
                if util.is_platform_file(os.path.join(dn, fn))
            ]
            real_names.sort()

            self.assertEqual(names, real_names)

        finally:
            sys.stderr = saved_stderr
            sys.argv = saved_argv

    def test_macho_find(self):
        fp = StringIO()
        macho_find.print_file(fp, "file1")
        macho_find.print_file(fp, "file2")
        self.assertEqual(fp.getvalue(), "file1\nfile2\n")

    def test_macho_dump(self):
        fp = StringIO()
        macho_dump.print_file(fp, "/bin/sh")
        lines = fp.getvalue().splitlines()

        self.assertEqual(lines[0], "/bin/sh")
        self.assertTrue(len(lines) > 3)

        self.assertEqual(lines[-1], "")
        del lines[-1]

        idx = 1
        while idx < len(lines):
            self.assertTrue(lines[idx].startswith("    [MachOHeader endian"))
            idx += 1

            lc = 0
            while idx < len(lines):
                if not lines[idx].startswith("\t"):
                    break

                lc += 1
                # self.assertTrue(os.path.exists(lines[idx].lstrip()))
                idx += 1

            self.assertTrue(lc > 0)


if __name__ == "__main__":
    unittest.main()



"""
Utilities for reading and writing Mach-O headers
"""

import os
import sys

from altgraph.ObjectGraph import ObjectGraph

from macholib.dyld import dyld_find
from macholib.itergraphreport import itergraphreport
from macholib.MachO import MachO

__all__ = ["MachOGraph"]

try:
    unicode
except NameError:
    unicode = str


class MissingMachO(object):
    def __init__(self, filename):
        self.graphident = filename
        self.headers = ()

    def __repr__(self):
        return "<%s graphident=%r>" % (type(self).__name__, self.graphident)


class MachOGraph(ObjectGraph):
    """
    Graph data structure of Mach-O dependencies
    """

    def __init__(self, debug=0, graph=None, env=None, executable_path=None):
        super(MachOGraph, self).__init__(debug=debug, graph=graph)
        self.env = env
        self.trans_table = {}
        self.executable_path = executable_path

    def locate(self, filename, loader=None):
        if not isinstance(filename, (str, unicode)):
            raise TypeError("%r is not a string" % (filename,))
        if filename.startswith("@loader_path/") and loader is not None:
            fn = self.trans_table.get((loader.filename, filename))
            if fn is None:
                loader_path = loader.loader_path

                try:
                    fn = dyld_find(
                        filename,
                        env=self.env,
                        executable_path=self.executable_path,
                        loader_path=loader_path,
                    )
                    self.trans_table[(loader.filename, filename)] = fn
                except ValueError:
                    return None

        else:
            fn = self.trans_table.get(filename)
            if fn is None:
                try:
                    fn = dyld_find(
                        filename, env=self.env, executable_path=self.executable_path
                    )
                    self.trans_table[filename] = fn
                except ValueError:
                    return None
        return fn

    def findNode(self, name, loader=None):
        assert isinstance(name, (str, unicode))
        data = super(MachOGraph, self).findNode(name)
        if data is not None:
            return data
        newname = self.locate(name, loader=loader)
        if newname is not None and newname != name:
            return self.findNode(newname)
        return None

    def run_file(self, pathname, caller=None):
        assert isinstance(pathname, (str, unicode))
        self.msgin(2, "run_file", pathname)
        m = self.findNode(pathname, loader=caller)
        if m is None:
            if not os.path.exists(pathname):
                raise ValueError("%r does not exist" % (pathname,))
            m = self.createNode(MachO, pathname)
            self.createReference(caller, m, edge_data="run_file")
            self.scan_node(m)
        self.msgout(2, "")
        return m

    def load_file(self, name, caller=None):
        assert isinstance(name, (str, unicode))
        self.msgin(2, "load_file", name, caller)
        m = self.findNode(name, loader=caller)
        if m is None:
            newname = self.locate(name, loader=caller)
            if newname is not None and newname != name:
                return self.load_file(newname, caller=caller)
            if os.path.exists(name):
                m = self.createNode(MachO, name)
                self.scan_node(m)
            else:
                m = self.createNode(MissingMachO, name)
        self.msgout(2, "")
        return m

    def scan_node(self, node):
        self.msgin(2, "scan_node", node)
        for header in node.headers:
            for _idx, name, filename in header.walkRelocatables():
                assert isinstance(name, (str, unicode))
                assert isinstance(filename, (str, unicode))
                m = self.load_file(filename, caller=node)
                self.createReference(node, m, edge_data=name)
        self.msgout(2, "", node)

    def itergraphreport(self, name="G"):
        nodes = map(self.graph.describe_node, self.graph.iterdfs(self))
        describe_edge = self.graph.describe_edge
        return itergraphreport(nodes, describe_edge, name=name)

    def graphreport(self, fileobj=None):
        if fileobj is None:
            fileobj = sys.stdout
        fileobj.writelines(self.itergraphreport())


def main(args):
    g = MachOGraph()
    for arg in args:
        g.run_file(arg)
    g.graphreport()


if __name__ == "__main__":
    main(sys.argv[1:] or ["/bin/ls"])



"""
Shared setup file for simple python packages. Uses a setup.cfg that
is the same as the distutils2 project, unless noted otherwise.

It exists for two reasons:
1) This makes it easier to reuse setup.py code between my own
   projects

2) Easier migration to distutils2 when that catches on.

Additional functionality:

* Section metadata:
    requires-test:  Same as 'tests_require' option for setuptools.

"""

import os
import platform
import re
import sys
from distutils import log
from fnmatch import fnmatch

from setuptools import Command, setup
from setuptools.command import egg_info

if sys.version_info[0] == 2:
    from ConfigParser import NoOptionError, NoSectionError, RawConfigParser
else:
    from configparser import NoOptionError, NoSectionError, RawConfigParser

ROOTDIR = os.path.dirname(os.path.abspath(__file__))


#
#
#
# Parsing the setup.cfg and converting it to something that can be
# used by setuptools.setup()
#
#
#


def eval_marker(value):
    """
    Evaluate an distutils2 environment marker.

    This code is unsafe when used with hostile setup.cfg files,
    but that's not a problem for our own files.
    """
    value = value.strip()

    class M:
        def __init__(self, **kwds):
            for k, v in kwds.items():
                setattr(self, k, v)

    variables = {
        "python_version": "%d.%d" % (sys.version_info[0], sys.version_info[1]),
        "python_full_version": sys.version.split()[0],
        "os": M(name=os.name),
        "sys": M(platform=sys.platform),
        "platform": M(version=platform.version(), machine=platform.machine()),
    }

    return bool(eval(value, variables, variables))

    return True


def _opt_value(cfg, into, section, key, transform=None):
    try:
        v = cfg.get(section, key)
        if transform != _as_lines and ";" in v:
            v, marker = v.rsplit(";", 1)
            if not eval_marker(marker):
                return

            v = v.strip()

        if v:
            if transform:
                into[key] = transform(v.strip())
            else:
                into[key] = v.strip()

    except (NoOptionError, NoSectionError):
        pass


def _as_bool(value):
    if value.lower() in ("y", "yes", "on"):
        return True
    elif value.lower() in ("n", "no", "off"):
        return False
    elif value.isdigit():
        return bool(int(value))
    else:
        raise ValueError(value)


def _as_list(value):
    return value.split()


def _as_lines(value):
    result = []
    for v in value.splitlines():
        if ";" in v:
            v, marker = v.rsplit(";", 1)
            if not eval_marker(marker):
                continue

            v = v.strip()
            if v:
                result.append(v)
        else:
            result.append(v)
    return result


def _map_requirement(value):
    m = re.search(r"(\S+)\s*(?:\((.*)\))?", value)
    name = m.group(1)
    version = m.group(2)

    if version is None:
        return name

    else:
        mapped = []
        for v in version.split(","):
            v = v.strip()
            if v[0].isdigit():
                # Checks for a specific version prefix
                m = v.rsplit(".", 1)
                mapped.append(">=%s,<%s.%s" % (v, m[0], int(m[1]) + 1))

            else:
                mapped.append(v)
        return "%s %s" % (name, ",".join(mapped))


def _as_requires(value):
    requires = []
    for req in value.splitlines():
        if ";" in req:
            req, marker = req.rsplit(";", 1)
            if not eval_marker(marker):
                continue
            req = req.strip()

        if not req:
            continue
        requires.append(_map_requirement(req))
    return requires


def parse_setup_cfg():
    cfg = RawConfigParser()
    r = cfg.read([os.path.join(ROOTDIR, "setup.cfg")])
    if len(r) != 1:
        print("Cannot read 'setup.cfg'")
        sys.exit(1)

    metadata = {
        "name": cfg.get("x-metadata", "name"),
        "version": cfg.get("x-metadata", "version"),
        "description": cfg.get("x-metadata", "description"),
    }

    _opt_value(cfg, metadata, "x-metadata", "license")
    _opt_value(cfg, metadata, "x-metadata", "maintainer")
    _opt_value(cfg, metadata, "x-metadata", "maintainer_email")
    _opt_value(cfg, metadata, "x-metadata", "author")
    _opt_value(cfg, metadata, "x-metadata", "author_email")
    _opt_value(cfg, metadata, "x-metadata", "url")
    _opt_value(cfg, metadata, "x-metadata", "download_url")
    _opt_value(cfg, metadata, "x-metadata", "classifiers", _as_lines)
    _opt_value(cfg, metadata, "x-metadata", "platforms", _as_list)
    _opt_value(cfg, metadata, "x-metadata", "packages", _as_list)
    _opt_value(cfg, metadata, "x-metadata", "keywords", _as_list)

    try:
        v = cfg.get("x-metadata", "requires-dist")

    except (NoOptionError, NoSectionError):
        pass

    else:
        requires = _as_requires(v)
        if requires:
            metadata["install_requires"] = requires

    try:
        v = cfg.get("x-metadata", "requires-test")

    except (NoOptionError, NoSectionError):
        pass

    else:
        requires = _as_requires(v)
        if requires:
            metadata["tests_require"] = requires

    try:
        v = cfg.get("x-metadata", "long_description_file")
    except (NoOptionError, NoSectionError):
        pass

    else:
        parts = []
        for nm in v.split():
            fp = open(nm, "r")
            parts.append(fp.read())
            fp.close()

        metadata["long_description"] = "\n\n".join(parts)
        metadata["long_description_content_type"] = "text/x-rst; charset=UTF-8"

    try:
        v = cfg.get("x-metadata", "zip-safe")
    except (NoOptionError, NoSectionError):
        pass

    else:
        metadata["zip_safe"] = _as_bool(v)

    try:
        v = cfg.get("x-metadata", "console_scripts")
    except (NoOptionError, NoSectionError):
        pass

    else:
        if "entry_points" not in metadata:
            metadata["entry_points"] = {}

        metadata["entry_points"]["console_scripts"] = v.splitlines()

    if sys.version_info[:2] <= (2, 6):
        try:
            metadata["tests_require"] += ", unittest2"
        except KeyError:
            metadata["tests_require"] = "unittest2"

    return metadata


#
#
#
# Definitions of custom commands
#
#
#


def recursiveGlob(root, pathPattern):
    """
    Recursively look for files matching 'pathPattern'. Return a list
    of matching files/directories.
    """
    result = []

    for rootpath, _dirnames, filenames in os.walk(root):
        for fn in filenames:
            if fnmatch(fn, pathPattern):
                result.append(os.path.join(rootpath, fn))
    return result


def importExternalTestCases(unittest, pathPattern="test_*.py", root=".", package=None):
    """
    Import all unittests in the PyObjC tree starting at 'root'
    """

    testFiles = recursiveGlob(root, pathPattern)
    testModules = [
        x[len(root) + 1 : -3].replace("/", ".") for x in testFiles  # noqa: E203
    ]  # noqa: E203
    if package is not None:
        testModules = [(package + "." + m) for m in testModules]

    suites = []

    for modName in testModules:
        try:
            module = __import__(modName)
        except ImportError:
            print("SKIP %s: %s" % (modName, sys.exc_info()[1]))
            continue

        if "." in modName:
            for elem in modName.split(".")[1:]:
                module = getattr(module, elem)

        s = unittest.defaultTestLoader.loadTestsFromModule(module)
        suites.append(s)

    return unittest.TestSuite(suites)


class my_egg_info(egg_info.egg_info):
    def run(self):
        egg_info.egg_info.run(self)

        path = os.path.join(self.egg_info, "PKG-INFO")

        with open(path, "r") as fp:
            contents = fp.read()

        try:
            before, after = contents.split("\n\n", 1)
        except ValueError:
            before = contents
            after = "\n\n"

        with open(path, "w") as fp:
            fp.write(before)
            fp.write(
                "\nProject-URL: Documentation, https://macholib.readthedocs.io/en/latest/\n"  # noqa: B950
            )
            fp.write(
                "Project-URL: Issue tracker, https://github.com/ronaldoussoren/macholib/issues\n"  # noqa: B950
            )
            fp.write(
                "Project-URL: Repository, https://github.com/ronaldoussoren/macholib\n\n"  # noqa: B950
            )
            fp.write(after)


class my_test(Command):
    description = "run test suite"
    user_options = [("verbosity=", None, "print what tests are run")]

    def initialize_options(self):
        self.verbosity = "1"

    def finalize_options(self):
        if isinstance(self.verbosity, str):
            self.verbosity = int(self.verbosity)

    def cleanup_environment(self):
        ei_cmd = self.get_finalized_command("egg_info")
        egg_name = ei_cmd.egg_name.replace("-", "_")

        to_remove = []
        for dirname in sys.path:
            bn = os.path.basename(dirname)
            if bn.startswith(egg_name + "-"):
                to_remove.append(dirname)

        for dirname in to_remove:
            log.info("removing installed %r from sys.path before testing" % (dirname,))
            sys.path.remove(dirname)

    def add_project_to_sys_path(self):
        from pkg_resources import (
            add_activation_listener,
            normalize_path,
            require,
            working_set,
        )

        self.reinitialize_command("egg_info")
        self.run_command("egg_info")
        self.reinitialize_command("build_ext", inplace=1)
        self.run_command("build_ext")

        # Check if this distribution is already on sys.path
        # and remove that version, this ensures that the right
        # copy of the package gets tested.

        self.__old_path = sys.path[:]
        self.__old_modules = sys.modules.copy()

        ei_cmd = self.get_finalized_command("egg_info")
        sys.path.insert(0, normalize_path(ei_cmd.egg_base))
        sys.path.insert(1, os.path.dirname(__file__))

        # Strip the namespace packages defined in this distribution
        # from sys.modules, needed to reset the search path for
        # those modules.

        nspkgs = getattr(self.distribution, "namespace_packages", None)
        if nspkgs is not None:
            for nm in nspkgs:
                del sys.modules[nm]

        # Reset pkg_resources state:
        add_activation_listener(lambda dist: dist.activate())
        working_set.__init__()
        require("%s==%s" % (ei_cmd.egg_name, ei_cmd.egg_version))

    def remove_from_sys_path(self):
        from pkg_resources import working_set

        sys.path[:] = self.__old_path
        sys.modules.clear()
        sys.modules.update(self.__old_modules)
        working_set.__init__()

    def run(self):
        import unittest

        # Ensure that build directory is on sys.path (py3k)

        self.cleanup_environment()
        self.add_project_to_sys_path()

        try:
            meta = self.distribution.metadata
            name = meta.get_name()
            test_pkg = name + "_tests"
            suite = importExternalTestCases(unittest, "test_*.py", test_pkg, test_pkg)

            runner = unittest.TextTestRunner(verbosity=self.verbosity)
            result = runner.run(suite)

            # Print out summary. This is a structured format that
            # should make it easy to use this information in scripts.
            summary = {
                "count": result.testsRun,
                "fails": len(result.failures),
                "errors": len(result.errors),
                "xfails": len(getattr(result, "expectedFailures", [])),
                "xpass": len(getattr(result, "expectedSuccesses", [])),
                "skip": len(getattr(result, "skipped", [])),
            }
            print("SUMMARY: %s" % (summary,))
            if summary["fails"] or summary["errors"]:
                sys.exit(1)

        finally:
            self.remove_from_sys_path()


#
#
#
#  And finally run the setuptools main entry point.
#
#
#

metadata = parse_setup_cfg()

setup(cmdclass={"test": my_test, "egg_info": my_egg_info}, **metadata)
